import { sumSerializerFixedSizes, mergeBytes, isSome, none, some, publicKey } from '@lorisleiva/js-core';
import { toWeb3JsPublicKey, fromWeb3JsPublicKey } from '@lorisleiva/js-web3js-adapters';
import * as beet from '@metaplex-foundation/beet';
import * as beetSolana from '@metaplex-foundation/beet-solana';
import { PublicKey } from '@solana/web3.js';
import { Buffer } from 'buffer';
import { OperationNotSupportedError } from './errors.mjs';
import { u32, u8, bool, u16, u64, u128, i8, i16, i32, i64, i128 } from './numbers.mjs';

class BeetSerializer {
  tuple(items, description) {
    const itemDescriptions = items.map(item => item.description).join(', ');
    return {
      description: description ?? `tuple(${itemDescriptions})`,
      fixedSize: sumSerializerFixedSizes(items),
      serialize: value => {
        if (value.length !== items.length) {
          throw new Error(`Expected tuple to have ${items.length} items but got ${value.length}.`);
        }
        return mergeBytes(items.map((item, index) => item.serialize(value[index])));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        items.forEach(serializer => {
          const [newValue, newOffset] = serializer.deserialize(bytes, offset);
          values.push(newValue);
          offset = newOffset;
        });
        return [values, offset];
      }
    };
  }
  vec(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? u32();
    return {
      description: description ?? `vec(${itemSerializer.description})`,
      fixedSize: null,
      serialize: value => {
        const lengthBytes = prefixSeralizer.serialize(value.length);
        const itemBytes = value.map(item => itemSerializer.serialize(item));
        return mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        const [length, newOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  }
  array(itemSerializer, size, description) {
    const childSize = itemSerializer.fixedSize;
    const fixedSize = childSize === null ? null : childSize * size;
    return {
      description: description ?? `array(${itemSerializer.description}; ${size})`,
      fixedSize: size === 0 ? 0 : fixedSize,
      serialize: value => {
        if (value.length !== size) {
          throw new Error(`Expected array to have ${size} items but got ${value.length}.`);
        }
        return mergeBytes(value.map(item => itemSerializer.serialize(item)));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        for (let i = 0; i < size; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  }
  map(keySerializer, valueSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? u32();
    return {
      description: description ?? `map(${keySerializer.description}, ${valueSerializer.description})`,
      fixedSize: null,
      serialize: map => {
        const lengthBytes = prefixSeralizer.serialize(map.size);
        const itemBytes = Array.from(map, ([key, value]) => mergeBytes([keySerializer.serialize(key), valueSerializer.serialize(value)]));
        return mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const map = new Map();
        const [length, newOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [key, kOffset] = keySerializer.deserialize(bytes, offset);
          offset = kOffset;
          const [value, vOffset] = valueSerializer.deserialize(bytes, offset);
          offset = vOffset;
          map.set(key, value);
        }
        return [map, offset];
      }
    };
  }
  set(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? u32();
    return {
      description: description ?? `set(${itemSerializer.description})`,
      fixedSize: null,
      serialize: set => {
        const lengthBytes = prefixSeralizer.serialize(set.size);
        const itemBytes = Array.from(set, value => itemSerializer.serialize(value));
        return mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const set = new Set();
        const [length, newOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          offset = newOffset;
          set.add(value);
        }
        return [set, offset];
      }
    };
  }
  option(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? u8();
    return {
      description: description ?? `option(${itemSerializer.description})`,
      fixedSize: itemSerializer.fixedSize !== null && prefixSeralizer.fixedSize !== null ? itemSerializer.fixedSize + prefixSeralizer.fixedSize : null,
      serialize: option => {
        const prefixByte = prefixSeralizer.serialize(Number(isSome(option)));
        const itemBytes = isSome(option) ? itemSerializer.serialize(option.value) : new Uint8Array();
        return mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const [isSome, prefixOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [none(), offset];
        }
        const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
        offset = newOffset;
        return [some(value), offset];
      }
    };
  }
  nullable(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? u8();
    return {
      description: description ?? `nullable(${itemSerializer.description})`,
      fixedSize: itemSerializer.fixedSize !== null && prefixSeralizer.fixedSize !== null ? itemSerializer.fixedSize + prefixSeralizer.fixedSize : null,
      serialize: option => {
        const prefixByte = prefixSeralizer.serialize(Number(option !== null));
        const itemBytes = option !== null ? itemSerializer.serialize(option) : new Uint8Array();
        return mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const [isSome, prefixOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [null, offset];
        }
        const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
        offset = newOffset;
        return [value, offset];
      }
    };
  }
  struct(fields, description) {
    const fieldDescriptions = fields.map(([name, serializer]) => `${String(name)}: ${serializer.description}`).join(', ');
    return {
      description: description ?? `struct(${fieldDescriptions})`,
      fixedSize: sumSerializerFixedSizes(fields.map(([, s]) => s)),
      serialize: struct => {
        const fieldBytes = fields.map(([key, serializer]) => serializer.serialize(struct[key]));
        return mergeBytes(fieldBytes);
      },
      deserialize: (bytes, offset = 0) => {
        const struct = {};
        fields.forEach(([key, serializer]) => {
          const [value, newOffset] = serializer.deserialize(bytes, offset);
          offset = newOffset;
          struct[key] = value;
        });
        return [struct, offset];
      }
    };
  }
  enum(constructor, description) {
    const enumValues = Object.values(constructor);
    const isNumericEnum = enumValues.some(v => typeof v === 'number');
    const valueDescriptions = enumValues.filter(v => typeof v === 'string').join(', ');
    function getVariantKeyValue(value) {
      if (typeof value === 'number') {
        return [enumValues[value], value];
      }
      const variantValue = constructor[value];
      if (typeof variantValue === 'number') {
        return [value, variantValue];
      }
      const indexOfValue = enumValues.indexOf(variantValue);
      if (indexOfValue >= 0) {
        return [variantValue, indexOfValue];
      }
      return [value, enumValues.indexOf(value)];
    }
    function checkVariantExists(variantKey) {
      if (!enumValues.includes(variantKey)) {
        throw new Error(`Invalid enum variant. Got "${variantKey}", expected one of ` + `[${enumValues.join(', ')}]`);
      }
    }
    return {
      description: description ?? `enum(${valueDescriptions})`,
      fixedSize: 1,
      serialize: value => {
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return u8().serialize(variantValue);
      },
      deserialize: (bytes, offset = 0) => {
        const [value, newOffset] = u8().deserialize(bytes, offset);
        offset = newOffset;
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return [isNumericEnum ? variantValue : variantKey, offset];
      }
    };
  }
  dataEnum(fields, prefix, description) {
    const prefixSeralizer = prefix ?? u8();
    const fieldDescriptions = fields.map(([name, serializer]) => `${String(name)}${serializer ? `: ${serializer.description}` : ''}`).join(', ');
    const allVariantHaveTheSameFixedSize = fields.every((one, i, all) => one[1].fixedSize === all[0][1].fixedSize);
    return {
      description: description ?? `dataEnum(${fieldDescriptions})`,
      fixedSize: allVariantHaveTheSameFixedSize && fields.length > 0 && fields[0][1].fixedSize !== null && prefixSeralizer.fixedSize !== null ? fields[0][1].fixedSize + prefixSeralizer.fixedSize : null,
      serialize: variant => {
        const discriminator = fields.findIndex(([key]) => variant.__kind === key);
        if (discriminator < 0) {
          throw new Error(`Invalid data enum variant. Got "${variant.__kind}", expected one of ` + `[${fields.map(([key]) => key).join(', ')}]`);
        }
        const variantPrefix = prefixSeralizer.serialize(discriminator);
        const variantSerializer = fields[discriminator][1];
        const variantBytes = variantSerializer.serialize(variant);
        return mergeBytes([variantPrefix, variantBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const [discriminator, dOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = dOffset;
        const variantField = fields[Number(discriminator)] ?? null;
        if (!variantField) {
          throw new Error(`Data enum index "${discriminator}" is out of range. ` + `Index should be between 0 and ${fields.length - 1}.`);
        }
        const [variant, vOffset] = variantField[1].deserialize(bytes, offset);
        offset = vOffset;
        return [{
          __kind: variantField[0],
          ...(variant ?? {})
        }, offset];
      }
    };
  }
  get unit() {
    return {
      description: 'unit',
      fixedSize: 0,
      serialize: () => new Uint8Array(),
      deserialize: (_bytes, offset = 0) => [undefined, offset]
    };
  }
  get bool() {
    return bool();
  }
  get u8() {
    return u8();
  }
  get u16() {
    return u16();
  }
  get u32() {
    return u32();
  }
  get u64() {
    return u64();
  }
  get u128() {
    return u128();
  }
  get i8() {
    return i8();
  }
  get i16() {
    return i16();
  }
  get i32() {
    return i32();
  }
  get i64() {
    return i64();
  }
  get i128() {
    return i128();
  }
  get f32() {
    return {
      description: 'f32 [not supported]',
      fixedSize: 4,
      serialize: () => {
        throw new OperationNotSupportedError('f32');
      },
      deserialize: () => {
        throw new OperationNotSupportedError('f32');
      }
    };
  }
  get f64() {
    return {
      description: 'f64 [not supported]',
      fixedSize: 8,
      serialize: () => {
        throw new OperationNotSupportedError('f64');
      },
      deserialize: () => {
        throw new OperationNotSupportedError('f64');
      }
    };
  }
  get string() {
    return {
      description: 'string',
      fixedSize: null,
      serialize: value => {
        const stringBeet = beet.utf8String.toFixedFromValue(value);
        const buffer = Buffer.alloc(stringBeet.byteSize);
        stringBeet.write(buffer, 0, value);
        return new Uint8Array(buffer);
      },
      deserialize: (bytes, offset = 0) => {
        const buffer = Buffer.from(bytes);
        const stringBeet = beet.utf8String.toFixedFromData(buffer, offset);
        const value = stringBeet.read(buffer, offset);
        return [value, offset + stringBeet.byteSize];
      }
    };
  }
  get bytes() {
    return {
      description: 'bytes',
      fixedSize: null,
      serialize: value => new Uint8Array(value),
      deserialize: (bytes, offset = 0) => [new Uint8Array(bytes), offset + bytes.length]
    };
  }
  get publicKey() {
    return {
      description: 'publicKey',
      fixedSize: 32,
      serialize: value => {
        const buffer = Buffer.alloc(beetSolana.publicKey.byteSize);
        const key = new PublicKey(toWeb3JsPublicKey(publicKey(value)));
        beetSolana.publicKey.write(buffer, 0, key);
        return new Uint8Array(buffer);
      },
      deserialize: (bytes, offset = 0) => {
        const buffer = Buffer.from(bytes);
        const value = beetSolana.publicKey.read(buffer, offset);
        return [fromWeb3JsPublicKey(value), offset + beetSolana.publicKey.byteSize];
      }
    };
  }
}

export { BeetSerializer };
//# sourceMappingURL=BeetSerializer.mjs.map
